from datetime import datetime, timedelta, date, time

from oldschool_management_tools.special_prompts.special_prompts import SPECIAL_PROMPTS
from oldschool_management_tools.calendar_items.calendar_item import CalendarItem
from oldschool_management_tools.calendar_integrator import CalendarIntegrator
from os import system
from dataclasses import dataclass
from re import match as re_match
import yaml

# Ideas:
#  - See allocated time per day
#  - Plan in breaks
#  - Draw schedule in hours

# Do the work where it is fast :)

CATEGORIES_REQUIRING_PREP = ["green"]
system('color')


class CalendarReader:

    def __init__(self, integrator: CalendarIntegrator):
        self.cal_integrator = integrator

        with open("management_tools_config.yaml", "r") as stream:
            config = yaml.safe_load(stream)
            self.config = config
            
    def get_day_cal(self, cal_date):
        return self.cal_integrator.get_day_cal(cal_date)

    def fill_prep(self, parsed_day=datetime.today()):
        prep_duration = self.cal_integrator.PREP_DURATION
        apts = self.get_valid_apts(self.cal_integrator.get_day_cal(parsed_day))
        free_slots: list[TimeSlot] = []
        last_apt_end_time: datetime = None
        new_apts = []

        for apt in apts:
            if not free_slots:
                free_slots.append(TimeSlot(apt.start_datetime.replace(hour=8, minute=0, second=0), apt.start_datetime))
            elif last_apt_end_time and free_slots[-1].end_datetime < apt.start_datetime and last_apt_end_time < apt.start_datetime:
                new_free_slot = TimeSlot(last_apt_end_time, apt.start_datetime)
                free_slots.append(new_free_slot)

            if apt.category.lower() in CATEGORIES_REQUIRING_PREP:
                last_free_slot = free_slots[-1]
                if last_free_slot.end_datetime < apt.end_datetime:
                    prep_slot = TimeSlot(last_free_slot.end_datetime - timedelta(minutes=prep_duration),
                                         last_free_slot.end_datetime)
                    last_free_slot.end_datetime = last_free_slot.end_datetime - timedelta(minutes=prep_duration)
                    if last_free_slot.end_datetime - timedelta(minutes=prep_duration) <= last_free_slot.start_datetime:
                        free_slots.pop()
                    new_apts.append(CalendarItem(f"Prep {apt.name}", "Desk", "yellow",
                                                 prep_slot.start_datetime, prep_slot.start_datetime.time(),
                                                 prep_slot.end_datetime, prep_slot.end_datetime.time()))
            last_apt_end_time = apt.end_datetime

        [self.cal_integrator.add_apt(apt) for apt in new_apts]

    def show_day_sched(self, parsed_day=datetime.today()):
        cal = self.get_day_cal(parsed_day)
        self.print_cal(cal)

    def wipe_prep(self, parsed_day=datetime.today()):
        cal = self.get_day_cal(parsed_day)
        cal_item_num = 0
        try:
            while len(cal) > cal_item_num + 1:
                cal_item = cal[cal_item_num]

                if cal_item.category == "yellow" and cal_item.name.startswith("Prep"):
                    self.cal_integrator.delete_cal_item(parsed_day, cal_item)
                    cal = self.get_day_cal(parsed_day)
                else:
                    cal_item_num = cal_item_num + 1
        except IndexError:
            # Probably fine
            pass

    def get_valid_apts(self, cal):
        return [apt for apt in cal if not re_match(self.config['ignore_subject_regex'], str(apt.name))]

    def print_cal(self, cal):
        for outlook_apt in self.get_valid_apts(cal):
            outlook_apt.print()

    def prompt_day_tasks(self, parsed_day=datetime.today()):
        for spec_prompt in SPECIAL_PROMPTS:
            spec_prompt.show()
        apts = self.get_valid_apts(self.get_day_cal(parsed_day))
        for apt in apts:
            if apt.category.lower() in CATEGORIES_REQUIRING_PREP:
                input(f"Prep for {apt.name} => Done?")


def start_of_day(day: date) -> datetime:
    return datetime.combine(day, time())


def parse_day(day) -> datetime:
    match day:
        case None | '' | "today":
            return start_of_day(date.today())
        case "tomorrow":
            return start_of_day(date.today()) + timedelta(days=1)
        case d if day.isnumeric():
            return start_of_day(date.today()) + timedelta(days=int(d))
        case _:
            raise ValueError(f"Bad day [{day}]")


@dataclass
class TimeSlot:
    start_datetime: datetime
    end_datetime: datetime



import dataclasses
from oldschool_management_tools.calendar_items.calendar_item import CalendarItem, RELEVANT_CAL_FIELDS
from datetime import datetime, timedelta, date, time
import os
import json
import pandas as pd
if os.name == "nt":
    from win32com import client


class CalendarIntegrator:

    def __init__(self):
        self.PREP_DURATION = 15
    
    def get_calendar(self, begin, end):
        pass
    
    def get_day_cal(self, cal_date: datetime):
        pass
    
    def add_apt(self, apt: CalendarItem):
        pass

    def delete_cal_item(self, cal_date: datetime, apt: CalendarItem):
        pass

    def get_day_cal_items(self, cal_date: datetime):
        return CalendarItem.from_outlook_apt(self.get_day_cal(cal_date))


class OutlookIntegrator(CalendarIntegrator):
    
    def __init__(self):
        super().__init__()
        self.outlook = client.Dispatch('Outlook.Application')
    
    def get_calendar(self, begin, end):
        mapi = self.outlook.GetNamespace('MAPI')
        calendar = mapi.getDefaultFolder(9).Items
        calendar.IncludeRecurrences = True
        calendar.Sort('[Start]')

        restriction = "[Start] >= '" + begin.strftime('%d/%m/%Y') + "' AND [END] <= '" + end.strftime(
            '%d/%m/%Y') + "'"
        calendar = calendar.Restrict(restriction)
        return calendar

    def get_day_cal(self, cal_date: datetime):
        apts = self.get_calendar(cal_date, cal_date + timedelta(days=1))
        return [CalendarItem.from_outlook_apt(apt) for apt in apts]

    def add_apt(self, apt: CalendarItem):
        apt.print()
        appt = self.outlook.CreateItem(1)
        appt.Start = str(apt.start_datetime)[0:-6]
        appt.Subject = apt.name
        appt.Categories = apt.category + " Category"
        appt.Duration = self.PREP_DURATION
        appt.Location = apt.location
        appt.Save()
        return appt

    def delete_cal_item(self, cal_date: datetime, apt: CalendarItem):
        outl_apts = self.get_calendar(cal_date, cal_date + timedelta(days=1))
        for outl_apt in outl_apts:
            if CalendarItem.from_outlook_apt(outl_apt) == apt:
                outl_apt.Delete()
                return


class DumpingOutlookIntegrator(OutlookIntegrator):

    def __init__(self, dump_dir):
        super().__init__()
        self.dump_dir = dump_dir
        self.dump_index = 0
        os.mkdir(dump_dir)

    def get_day_cal(self, cal_date: datetime):
        result = super().get_day_cal(cal_date)
        self.dump_items(result, "get_day_cal", [str(cal_date)])
        return result

    def add_apt(self, apt: CalendarItem):
        result = super().add_apt(apt)
        self.dump_items([result], "add_apt", [apt.to_json()])
        return result

    def dump_items(self, calendar, function_called, args):
        items = [CalendarItem.from_outlook_apt(i).to_dict() for i in calendar]
        df = pd.DataFrame(data=items)
        df.to_parquet(f"{self.dump_dir}/{function_called}-{self.dump_index}.parquet")
        self.dump_index += 1

    def delete_cal_item(self, cal_date: datetime, apt: CalendarItem):
        super().delete_cal_item(cal_date, apt)


class ParquetIntegrator(CalendarIntegrator):

    def __init__(self, input_file):
        super().__init__()
        self.df = pd.read_parquet(input_file)
        self.items = self.df.apply(lambda row: CalendarItem(*row), axis=1).tolist()

    def get_calendar(self, begin, end):
        return self.items

    def get_day_cal(self, cal_date: datetime):
        return self.get_calendar(cal_date, cal_date + timedelta(days=1))

    def add_apt(self, apt: CalendarItem):
        self.items.append(apt)
        self.items.sort(key=lambda x: x.start_datetime)
        return

    def delete_cal_item(self, cal_date: datetime, apt: CalendarItem):
        self.items.remove(apt)



import cmd
from sys import argv
from oldschool_management_tools.calendar_tools import parse_day
from oldschool_management_tools.calendar_tools import CalendarReader
from oldschool_management_tools.calendar_integrator import CalendarIntegrator, OutlookIntegrator, ParquetIntegrator, DumpingOutlookIntegrator


class MgmtToolsCmd(cmd.Cmd):
    intro = "Welcome to Old School Management Tools"
    prompt = "(mgmt) "

    def __init__(self, integrator: CalendarIntegrator):
        super(MgmtToolsCmd, self).__init__()
        self.calendar_reader = CalendarReader(integrator)

    def do_prompt_tasks(self, day):
        parsed_day = parse_day(day)
        self.calendar_reader.prompt_day_tasks(parsed_day)

    def do_show_sched(self, day):
        parsed_day = parse_day(day)
        self.calendar_reader.show_day_sched(parsed_day)

    def do_fill_prep(self, day):
        parsed_day = parse_day(day)
        self.calendar_reader.fill_prep(parsed_day)

    def do_wipe_prep(self, day):
        parsed_day = parse_day(day)
        self.calendar_reader.wipe_prep(parsed_day)

    @staticmethod
    def do_die(args):
        return True

    @staticmethod
    def do_exit(args):
        return True

    @staticmethod
    def do_EOF(args):
        return True


def run():
    if len(argv) == 1:
        integrator = OutlookIntegrator()
    elif argv[1] == "parquet":
        parq_file = argv[2]
        integrator = ParquetIntegrator(parq_file)
    elif argv[1] == "dump":
        integrator = DumpingOutlookIntegrator(argv[2])
    else:
        print(f"Bad input {argv[1]}")
        exit(1)
    MgmtToolsCmd(integrator).cmdloop()
    exit(0)
